"use strict";
/**
 * Expressive TypeScript Logger for Node.js
 * @packageDocumentation
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.Logger = void 0;
const util_1 = require("util");
const os_1 = require("os");
const path_1 = require("path");
const source_map_support_1 = require("source-map-support");
const LoggerHelper_1 = require("./LoggerHelper");
/**
 * 📝 Expressive TypeScript Logger for Node.js
 * @public
 */
class Logger {
    /**
     * @param settings - Configuration of the logger instance  (all settings are optional with sane defaults)
     */
    constructor(settings) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r, _s;
        this._logLevels = [
            "silly",
            "trace",
            "debug",
            "info",
            "warn",
            "error",
            "fatal",
        ];
        this._ignoreStackLevels = 3;
        this._attachedTransports = [];
        this._minLevelToStdErr = 4;
        const displayInstanceName = (settings === null || settings === void 0 ? void 0 : settings.displayInstanceName) === true;
        const setCallerAsLoggerName = (settings === null || settings === void 0 ? void 0 : settings.setCallerAsLoggerName) === true;
        this.settings = {
            type: (_a = settings === null || settings === void 0 ? void 0 : settings.type) !== null && _a !== void 0 ? _a : "pretty",
            displayInstanceName: displayInstanceName,
            instanceName: displayInstanceName
                ? (_b = settings === null || settings === void 0 ? void 0 : settings.instanceName) !== null && _b !== void 0 ? _b : os_1.hostname() : undefined,
            name: (_c = settings === null || settings === void 0 ? void 0 : settings.name) !== null && _c !== void 0 ? _c : (setCallerAsLoggerName
                ? (_e = (_d = LoggerHelper_1.LoggerHelper.getCallSites()[0].getTypeName()) !== null && _d !== void 0 ? _d : LoggerHelper_1.LoggerHelper.getCallSites()[0].getFunctionName()) !== null && _e !== void 0 ? _e : undefined : undefined),
            setCallerAsLoggerName: setCallerAsLoggerName,
            minLevel: (_f = settings === null || settings === void 0 ? void 0 : settings.minLevel) !== null && _f !== void 0 ? _f : "silly",
            exposeStack: (_g = settings === null || settings === void 0 ? void 0 : settings.exposeStack) !== null && _g !== void 0 ? _g : false,
            exposeErrorCodeFrame: (_h = settings === null || settings === void 0 ? void 0 : settings.exposeErrorCodeFrame) !== null && _h !== void 0 ? _h : true,
            exposeErrorCodeFrameLinesBeforeAndAfter: (_j = settings === null || settings === void 0 ? void 0 : settings.exposeErrorCodeFrameLinesBeforeAndAfter) !== null && _j !== void 0 ? _j : 5,
            suppressStdOutput: (_k = settings === null || settings === void 0 ? void 0 : settings.suppressStdOutput) !== null && _k !== void 0 ? _k : false,
            overwriteConsole: (_l = settings === null || settings === void 0 ? void 0 : settings.overwriteConsole) !== null && _l !== void 0 ? _l : false,
            logLevelsColors: (_m = settings === null || settings === void 0 ? void 0 : settings.logLevelsColors) !== null && _m !== void 0 ? _m : {
                0: "whiteBright",
                1: "white",
                2: "greenBright",
                3: "blueBright",
                4: "yellowBright",
                5: "redBright",
                6: "magentaBright",
            },
            prettyInspectHighlightStyles: {
                special: "cyan",
                number: "green",
                bigint: "green",
                boolean: "yellow",
                undefined: "red",
                null: "red",
                string: "red",
                symbol: "green",
                date: "magenta",
                name: "white",
                regexp: "red",
                module: "underline",
                ...settings === null || settings === void 0 ? void 0 : settings.prettyInspectHighlightStyles,
            },
            prettyInspectOptions: (_p = (_o = { compact: false, depth: Infinity, colors: true }) !== null && _o !== void 0 ? _o : settings === null || settings === void 0 ? void 0 : settings.prettyInspectOptions) !== null && _p !== void 0 ? _p : {
                colors: true,
                compact: false,
                depth: Infinity,
            },
            jsonInspectOptions: (_q = settings === null || settings === void 0 ? void 0 : settings.jsonInspectOptions) !== null && _q !== void 0 ? _q : {
                colors: false,
                compact: true,
                depth: Infinity,
            },
            stdOut: (_r = settings === null || settings === void 0 ? void 0 : settings.stdOut) !== null && _r !== void 0 ? _r : process.stdout,
            stdErr: (_s = settings === null || settings === void 0 ? void 0 : settings.stdErr) !== null && _s !== void 0 ? _s : process.stderr,
        };
        LoggerHelper_1.LoggerHelper.setUtilsInspectStyles(this.settings.prettyInspectHighlightStyles);
        LoggerHelper_1.LoggerHelper.initErrorToJsonHelper();
        if (this.settings.overwriteConsole) {
            LoggerHelper_1.LoggerHelper.overwriteConsole(this, this._handleLog);
        }
    }
    /**
     *  Attaches external Loggers, e.g. external log services, file system, database
     *
     * @param transportLogger - External logger to be attached. Must implement all log methods.
     * @param minLevel        - Minimum log level to be forwarded to this attached transport logger. (e.g. debug)
     */
    attachTransport(transportLogger, minLevel = "silly") {
        this._attachedTransports.push({
            minLevel,
            transportLogger,
        });
    }
    /**
     * Logs a silly message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    silly(...args) {
        return this._handleLog.apply(this, ["silly", args]);
    }
    /**
     * Logs a trace message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    trace(...args) {
        return this._handleLog.apply(this, ["trace", args, true]);
    }
    /**
     * Logs a debug message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    debug(...args) {
        return this._handleLog.apply(this, ["debug", args]);
    }
    /**
     * Logs an info message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    info(...args) {
        return this._handleLog.apply(this, ["info", args]);
    }
    /**
     * Logs a warn message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    warn(...args) {
        return this._handleLog.apply(this, ["warn", args]);
    }
    /**
     * Logs an error message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    error(...args) {
        return this._handleLog.apply(this, ["error", args]);
    }
    /**
     * Logs a fatal message.
     * @param args  - Multiple log attributes that should be logged out.
     */
    fatal(...args) {
        return this._handleLog.apply(this, ["fatal", args]);
    }
    /**
     * Helper: Pretty print error without logging it
     * @param error - Error object
     * @param print - Print the error or return only? (default: true)
     * @param exposeErrorCodeFrame  - Should the code frame be exposed? (default: true)
     * @param exposeStackTrace  - Should the stack trace be exposed? (default: true)
     * @param stackOffset - Offset lines of the stack trace (default: 0)
     * @param stackLimit  - Limit number of lines of the stack trace (default: Infinity)
     * @param std - Which std should the output be printed to? (default: stdErr)
     */
    prettyError(error, print = true, exposeErrorCodeFrame = true, exposeStackTrace = true, stackOffset = 0, stackLimit = Infinity, std = this.settings.stdErr) {
        const errorObject = this._buildErrorObject(error, exposeErrorCodeFrame, stackOffset, stackLimit);
        if (print) {
            this._printPrettyError(std, errorObject, exposeStackTrace);
        }
        return errorObject;
    }
    _handleLog(logLevel, logArguments, exposeStack = this.settings.exposeStack) {
        const logObject = this._buildLogObject(logLevel, logArguments, exposeStack);
        if (!this.settings.suppressStdOutput &&
            logObject.logLevelId >= this._logLevels.indexOf(this.settings.minLevel)) {
            if (this.settings.type === "pretty") {
                this._printPrettyLog(logObject);
            }
            else {
                this._printJsonLog(logObject);
            }
        }
        this._attachedTransports.forEach((transport) => {
            if (logObject.logLevelId >=
                Object.values(this._logLevels).indexOf(this.settings.minLevel)) {
                transport.transportLogger[logLevel](logObject);
            }
        });
        return logObject;
    }
    _buildLogObject(logLevel, logArguments, exposeStack = true) {
        const callSites = LoggerHelper_1.LoggerHelper.getCallSites();
        const relevantCallSites = callSites.splice(this._ignoreStackLevels);
        const stackFrame = source_map_support_1.wrapCallSite(relevantCallSites[0]);
        const stackFrameObject = LoggerHelper_1.LoggerHelper.toStackFrameObject(stackFrame);
        const logObject = {
            instanceName: this.settings.instanceName,
            loggerName: this.settings.name,
            hostname: os_1.hostname(),
            date: new Date(),
            logLevel: logLevel,
            logLevelId: this._logLevels.indexOf(logLevel),
            filePath: stackFrameObject.filePath,
            fullFilePath: stackFrameObject.fullFilePath,
            fileName: stackFrameObject.fileName,
            lineNumber: stackFrameObject.lineNumber,
            columnNumber: stackFrameObject.columnNumber,
            isConstructor: stackFrameObject.isConstructor,
            functionName: stackFrameObject.functionName,
            typeName: stackFrameObject.typeName,
            methodName: stackFrameObject.methodName,
            argumentsArray: [],
        };
        logArguments.forEach((arg) => {
            if (arg != null && typeof arg === "object" && LoggerHelper_1.LoggerHelper.isError(arg)) {
                logObject.argumentsArray.push(this._buildErrorObject(arg, this.settings.exposeErrorCodeFrame));
            }
            else {
                logObject.argumentsArray.push(arg);
            }
        });
        if (exposeStack) {
            logObject.stack = this._toStackObjectArray(relevantCallSites);
        }
        return logObject;
    }
    _buildErrorObject(error, exposeErrorCodeFrame = true, stackOffset = 0, stackLimit = Infinity) {
        var _a;
        const errorCallSites = LoggerHelper_1.LoggerHelper.getCallSites(error);
        stackOffset = stackOffset > -1 ? stackOffset : 0;
        const relevantCallSites = errorCallSites.splice(stackOffset);
        stackLimit = stackLimit > -1 ? stackLimit : 0;
        if (stackLimit < Infinity) {
            relevantCallSites.length = stackLimit;
        }
        const errorObject = JSON.parse(JSON.stringify(error));
        errorObject.nativeError = error;
        errorObject.details = { ...error };
        errorObject.name = (_a = errorObject.name) !== null && _a !== void 0 ? _a : "Error";
        errorObject.isError = true;
        errorObject.stack = this._toStackObjectArray(relevantCallSites);
        if (errorObject.stack.length > 0) {
            const errorCallSite = LoggerHelper_1.LoggerHelper.toStackFrameObject(source_map_support_1.wrapCallSite(relevantCallSites[0]));
            if (exposeErrorCodeFrame && errorCallSite.lineNumber != null) {
                if (errorCallSite.fullFilePath.indexOf("node_modules") < 0) {
                    errorObject.codeFrame = LoggerHelper_1.LoggerHelper._getCodeFrame(errorCallSite.fullFilePath, errorCallSite.lineNumber, errorCallSite.columnNumber, this.settings.exposeErrorCodeFrameLinesBeforeAndAfter);
                }
            }
        }
        return errorObject;
    }
    _toStackObjectArray(jsStack) {
        const stackFrame = Object.values(jsStack).reduce((stackFrameObj, callsite) => {
            stackFrameObj.push(LoggerHelper_1.LoggerHelper.toStackFrameObject(source_map_support_1.wrapCallSite(callsite)));
            return stackFrameObj;
        }, []);
        return stackFrame;
    }
    _printPrettyLog(logObject) {
        const std = logObject.logLevelId < this._minLevelToStdErr
            ? this.settings.stdOut
            : this.settings.stdErr;
        const nowStr = logObject.date
            .toISOString()
            .replace("T", " ")
            .replace("Z", "");
        const colorName = this.settings.logLevelsColors[logObject.logLevelId];
        std.write(LoggerHelper_1.LoggerHelper.styleString(["gray"], `${nowStr}\t`));
        std.write(LoggerHelper_1.LoggerHelper.styleString([colorName, "bold"], ` ${logObject.logLevel.toUpperCase()} `) + "\t");
        const functionName = logObject.isConstructor
            ? ` ${logObject.typeName}.constructor`
            : logObject.methodName != null
                ? ` ${logObject.typeName}.${logObject.methodName}`
                : logObject.functionName != null
                    ? ` ${logObject.functionName}`
                    : "";
        const instanceName = this.settings.instanceName != null
            ? `@${this.settings.instanceName}`
            : "";
        const loggerName = logObject.loggerName != null ? logObject.loggerName : "";
        const name = (loggerName + instanceName).length > 0
            ? `${loggerName}${instanceName} `
            : "";
        std.write(LoggerHelper_1.LoggerHelper.styleString(["gray"], `[${name}${logObject.filePath}:${logObject.lineNumber}${functionName}]`) + "  \t");
        logObject.argumentsArray.forEach((argument) => {
            const errorObject = argument;
            if (typeof argument === "object" && (errorObject === null || errorObject === void 0 ? void 0 : errorObject.isError) === true) {
                this._printPrettyError(std, errorObject);
            }
            else if (typeof argument === "object" &&
                (errorObject === null || errorObject === void 0 ? void 0 : errorObject.isError) !== true) {
                std.write("\n" + util_1.inspect(argument, this.settings.prettyInspectOptions));
            }
            else {
                std.write(util_1.format(argument) + " ");
            }
        });
        std.write("\n");
        if (logObject.stack != null) {
            std.write(LoggerHelper_1.LoggerHelper.styleString(["underline", "bold"], "log stack:\n"));
            this._printPrettyStack(std, logObject.stack);
        }
    }
    _printPrettyError(std, errorObject, printStackTrace = true) {
        std.write("\n" +
            LoggerHelper_1.LoggerHelper.styleString(["bgRed", "whiteBright", "bold"], ` ${errorObject.name} `) +
            `\t${util_1.format(errorObject.message)}`);
        if (Object.values(errorObject.details).length > 0) {
            std.write(LoggerHelper_1.LoggerHelper.styleString(["underline", "bold"], "\ndetails:"));
            std.write("\n" + util_1.inspect(errorObject.details, this.settings.prettyInspectOptions));
        }
        if (printStackTrace === true) {
            std.write(LoggerHelper_1.LoggerHelper.styleString(["underline", "bold"], "\nerror stack:"));
            this._printPrettyStack(std, errorObject.stack);
        }
        if (errorObject.codeFrame != null) {
            this._printPrettyCodeFrame(std, errorObject.codeFrame);
        }
    }
    _printPrettyStack(std, stackObjectArray) {
        std.write("\n");
        Object.values(stackObjectArray).forEach((stackObject) => {
            var _a;
            std.write(LoggerHelper_1.LoggerHelper.styleString(["gray"], "• ") +
                LoggerHelper_1.LoggerHelper.styleString(["yellowBright"], stackObject.fileName) +
                LoggerHelper_1.LoggerHelper.styleString(["gray"], ":") +
                LoggerHelper_1.LoggerHelper.styleString(["yellow"], stackObject.lineNumber) +
                LoggerHelper_1.LoggerHelper.styleString(["white"], " " + ((_a = stackObject.functionName) !== null && _a !== void 0 ? _a : "<anonymous>")));
            std.write("\n    ");
            std.write(path_1.normalize(LoggerHelper_1.LoggerHelper.styleString(["gray"], `${stackObject.filePath}:${stackObject.lineNumber}:${stackObject.columnNumber}`)));
            std.write("\n\n");
        });
    }
    _printPrettyCodeFrame(std, codeFrame) {
        std.write(LoggerHelper_1.LoggerHelper.styleString(["underline", "bold"], "code frame:\n"));
        let lineNumber = codeFrame.firstLineNumber;
        codeFrame.linesBefore.forEach((line) => {
            std.write(`  ${LoggerHelper_1.LoggerHelper.lineNumberTo3Char(lineNumber)} | ${line}\n`);
            lineNumber++;
        });
        std.write(LoggerHelper_1.LoggerHelper.styleString(["red"], ">") +
            " " +
            LoggerHelper_1.LoggerHelper.styleString(["bgRed", "whiteBright"], LoggerHelper_1.LoggerHelper.lineNumberTo3Char(lineNumber)) +
            " | " +
            LoggerHelper_1.LoggerHelper.styleString(["yellow"], codeFrame.relevantLine) +
            "\n");
        lineNumber++;
        if (codeFrame.columnNumber != null) {
            const positionMarker = new Array(codeFrame.columnNumber + 8).join(" ") + `^`;
            std.write(LoggerHelper_1.LoggerHelper.styleString(["red"], positionMarker) + "\n");
        }
        codeFrame.linesAfter.forEach((line) => {
            std.write(`  ${LoggerHelper_1.LoggerHelper.lineNumberTo3Char(lineNumber)} | ${line}\n`);
            lineNumber++;
        });
    }
    _printJsonLog(logObject) {
        const std = logObject.logLevelId < this._minLevelToStdErr
            ? this.settings.stdOut
            : this.settings.stdErr;
        const logObjectStringifiable = {
            ...logObject,
            argumentsArray: logObject.argumentsArray.map((argument) => {
                const errorObject = argument;
                if (typeof argument === "object" && errorObject.isError) {
                    let errorString;
                    try {
                        errorString = util_1.format(errorObject.nativeError);
                    }
                    catch (err) {
                        errorString = errorObject.nativeError.toString();
                    }
                    return {
                        ...errorObject,
                        nativeError: undefined,
                        errorString,
                    };
                }
                else {
                    return util_1.inspect(argument, this.settings.jsonInspectOptions);
                }
            }),
        };
        std.write(JSON.stringify(logObjectStringifiable) + "\n");
    }
}
exports.Logger = Logger;
//# sourceMappingURL=index.js.map